@@ -35,6 +35,7 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
     }
     checkTerms( data.tmb1$terms,   data.tmb0$terms )
     checkTerms( data.tmb1$termszi, data.tmb0$termszi )
+    checkTerms( data.tmb1$termsdisp, data.tmb0$termsdisp )
     ## Fixed effect parameters must be identical
     checkModelMatrix <- function(X1, X0) {
         if( !identical(colnames(X1), colnames(X0)) ) {
@@ -52,7 +53,7 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
     }
     checkModelMatrix(getX(data.tmb1), getX(data.tmb0))
     checkModelMatrix(getX(data.tmb1,"zi"), getX(data.tmb0,"zi"))
-    checkModelMatrix(getX(data.tmb1,"d"), getX(data.tmb0,"d"))
+    checkModelMatrix(getX(data.tmb1,"disp"), getX(data.tmb0,"disp"))
     NULL
 }
 
@@ -67,8 +68,11 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
 ##' \code{p} as the zero-inflation probability,
 ##' the possible choices are:
 ##' \describe{
-##' \item{"link"}{conditional mean on the scale of the link function,
-##' or equivalently the linear predictor of the conditional model}
+##' \item{"link"}{the linear predictor of the conditional model, or
+##' equivalently the conditional mean on the scale of the link function
+##' (this equivalence does not hold for truncated distributions, where
+##' the link-scaled value is not adjusted for the effect of truncation on the mean; to get the corrected value of the conditional mean on the linear predictor
+##' scale, use \code{family(m)$linkfun(predict(m, type = "conditional"))})}
 ##' \item{"response"}{expected value; this is \eqn{mu*(1-p)} for zero-inflated models
 ##' and \code{mu} otherwise}
 ##' \item{"conditional"}{mean of the conditional response; \code{mu} for all models
@@ -76,12 +80,16 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
 ##' \item{"zprob"}{the probability of a structural zero (returns 0 for non-zero-inflated models)}
 ##' \item{"zlink"}{predicted zero-inflation probability on the scale of
 ##' the logit link function (returns \code{-Inf} for non-zero-inflated models)}
-##' \item{"disp"}{dispersion parameter however it is defined for that particular family as described in  \code{\link{sigma.glmmTMB}}}
+##' \item{"disp"}{dispersion parameter, however it is defined for that particular family (as described in  \code{\link{sigma.glmmTMB}})}
+##' \item{"latent"}{return latent variables}
 ##' }
 ##' @param na.action how to handle missing values in \code{newdata} (see \code{\link{na.action}});
 ##' the default (\code{na.pass}) is to predict \code{NA}
 ##' @param debug (logical) return the \code{TMBStruc} object that will be
 ##' used internally for debugging?
+##' @param aggregate (optional factor vector) sum the elements with matching factor levels
+##' @param do.bias.correct (logical) should aggregated predictions use Taylor expanded estimate of nonlinear contribution of random effects (see details)
+##' @param bias.correct.control a list sent to TMB's function \code{sdreport()}. See documentation there.
 ##' @param re.form \code{NULL} to specify individual-level predictions; \code{~0} or \code{NA} to specify population-level predictions (i.e., setting all random effects to zero)
 ##' @param allow.new.levels allow previously unobserved levels in random-effects variables? see details.
 ##' @param \dots unused - for method compatibility
@@ -93,7 +101,11 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
 ##' However, to ensure intentional usage, a warning is triggered if \code{allow.new.levels=FALSE} (the default).
 
 ##' \item Prediction using "data-dependent bases" (variables whose scaling or transformation depends on the original data, e.g. \code{\link{poly}}, \code{\link[splines]{ns}}, or \code{\link{poly}}) should work properly; however, users are advised to check results extra-carefully when using such variables. Models with different versions of the same data-dependent basis type in different components (e.g. \code{formula= y ~ poly(x,3), dispformula= ~poly(x,2)}) will probably \emph{not} produce correct predictions.
+##' \item Bias corrected predictions are based on the method described in Thorson J.T. & Kristensen (2016). These should be checked carefully by the user and are not extensively tested.
 ##' }
+##' @references
+##' Thorson J.T. & Kristensen K. (2016) Implementing a generic method for bias correction in statistical models using random effects, with spatial and population dynamics examples. \emph{Fish. Res.} 175, 66-74. 
+##'
 ##'
 ##' @examples
 ##' data(sleepstudy,package="lme4")
@@ -107,6 +119,9 @@ assertIdenticalModels <- function(data.tmb1, data.tmb0, allow.new.levels=FALSE)
 ##' nd_pop <- data.frame(Days=unique(sleepstudy$Days),
 ##'                      Subject=NA)
 ##' predict(g0, newdata=nd_pop)
+##' ## return latent variables (BLUPs/conditional modes/etc. ) with standard errors
+##' ##  (actually conditional standard deviations)
+##' predict(g0, type = "latent", se.fit = TRUE)
 ##' @importFrom TMB sdreport
 ##' @importFrom stats optimHess model.frame na.fail na.pass napredict contrasts<-
 ##' @export
@@ -115,16 +130,20 @@ predict.glmmTMB <- function(object,
                             newparams=NULL,
                             se.fit=FALSE,
                             cov.fit=FALSE,
-                            re.form=NULL, allow.new.levels=FALSE,
+                            re.form=NULL,
+                            allow.new.levels=FALSE,
                             type = c("link", "response",
                                      "conditional", "zprob", "zlink",
-                                     "disp"),
+                                     "disp", "latent"),
                             zitype = NULL,
                             na.action = na.pass,
                             fast=NULL,
                             debug=FALSE,
+                            aggregate=NULL,
+                            do.bias.correct=FALSE,
+                            bias.correct.control = list(sd = TRUE),
                             ...) {
-  ## FIXME: add re.form
+  ## FIXME: implement 'complete' re.form (e.g. identify elements of Z or b that need to be zeroed out)
 
   check_dots(..., .action = "warning")
                
@@ -132,7 +151,17 @@ predict.glmmTMB <- function(object,
       if (!se.fit) message("se.fit set to TRUE because cov.fit = TRUE")
       se.fit <- TRUE
   }
-
+ 
+	if(is.null(aggregate)) {
+		aggregate <- factor()
+	}
+  ## FIXME: add re.form
+  if (length(aggregate) > 0) {
+    fast <- FALSE
+  }
+  if (do.bias.correct) {
+    se.fit <- TRUE
+  }
   if (!is.null(zitype)) {
      warning("zitype is deprecated: please use type instead")
      type <- zitype
@@ -156,6 +185,7 @@ predict.glmmTMB <- function(object,
                      zlink      = ,
                      zprob      = "prob",
                      disp       = "disp", #zi irrelevant; just reusing variable
+                     latent     = "uncorrected",  ## ignored, but needs to have a legal value
                      stop("unknown type ",type))
   ziPredCode <- .valid_zipredictcode[ziPredNm]
 
@@ -171,8 +201,13 @@ predict.glmmTMB <- function(object,
   if (is.null(fast)) fast <- !new_stuff
 
   ## what to ADREPORT:
-  ## 0 = no pred; 1 = response scale; 2 = link scale
-  do_pred_val <- if (!se.fit) 0 else if (!grepl("link",type)) 1 else 2
+  ## 0 = no pred; 1 = response scale; 2 = link scale; 3 = latent vars
+    do_pred_val <- if (!se.fit) {  0
+                   } else if (type == "latent") {
+                       3
+                   } else if (!grepl("link",type)) {
+                       1
+                   } else 2
 
   na.act <- attr(model.frame(object),"na.action")
   do.napred <- missing(newdata) && !is.null(na.act)
@@ -181,7 +216,7 @@ predict.glmmTMB <- function(object,
   ## want to make a giant if-block
   ## ('goto' would be handy here ...)
   if (noZI(object) && type %in% c("zprob", "zlink")) {
-    dd <- if (!is.null(newdata)) newdata else object$obj$env$data$Xd
+    dd <- if (!is.null(newdata)) newdata else object$obj$env$data$Xdisp
     pred <- se <- setNames(numeric(nrow(dd)), rownames(dd))
     se[] <- NA_real_
     pred[] <- if (type == "zprob") 0 else -Inf
@@ -348,12 +383,14 @@ predict.glmmTMB <- function(object,
                                fr=augFr,
                                yobs=yobs,
                                respCol=respCol,
-                               weights=model.weights(augFr),
+                               ## need to strip attributes
+                               weights=c(model.weights(augFr)),
                                contrasts=omi$contrasts,
                                family=omi$family,
                                ziPredictCode=ziPredNm,
                                doPredict=do_pred_val,
                                whichPredict=w,
+                               aggregate=aggregate,
                                REML=omi$REML,
                                map=omi$map,
                                sparseX=omi$sparseX,
@@ -363,7 +400,7 @@ predict.glmmTMB <- function(object,
                     )
 
     ## drop rank-deficient columns if necessary
-    for (nm in c("", "zi", "d")) {
+    for (nm in c("", "zi", "disp")) {
         xnm <- paste0("X", nm)
         betanm <- paste0("beta", nm)
         X <- getME(object, xnm)
@@ -387,7 +424,7 @@ predict.glmmTMB <- function(object,
                              any(!is.finite(Z@x)) |
                              any(!is.finite(Xzi)) |
                              any(!is.finite(Zzi@x)) |
-                             any(!is.finite(Xd))
+                             any(!is.finite(Xdisp))
     ) stop("Some variables in newdata needed for predictions contain NAs or NaNs.
            This is currently incompatible with se.fit=TRUE or cov.fit=TRUE."))
   }
@@ -405,17 +442,21 @@ predict.glmmTMB <- function(object,
   }
 
   if (pop_pred) {
+
+      ## use re.form, ll. 749ff of utils.R to decide which
+      ##  b values to set to zero.  OK to map _all_ values in this case
+      ##  (unless they're in newparams) ?
       TMBStruc <- within(TMBStruc, {
           parameters$b[] <- 0
           mapArg$b <- factor(rep(NA,length(parameters$b)))
       })
   }
 
-  n_orig <- openmp(n = object$modelInfo$parallel)
+  n_orig <- do.call(openmp, object$modelInfo$parallel)
   if (openmp_debug()) {
       cat("predict: setting OpenMP threads to ", n_orig, " on exit\n")
   }
-  on.exit(openmp(n_orig), add = TRUE)
+  on.exit(do.call(openmp, n_orig), add = TRUE)
 
   newObj <- with(TMBStruc,
                  MakeADFun(data.tmb,
@@ -431,18 +472,25 @@ predict.glmmTMB <- function(object,
   }  ## NOT fast
 
   ## set TMB threads to value from original model fit/reset on exit
-  if (!is.null(parallel <- object$modelInfo$parallel)) {
-    n_orig <- openmp(NULL)
-    if (openmp_debug()) cat("resetting TMB threads to ",  parallel, "\n")
-    openmp(parallel)
-    on.exit(openmp(n = n_orig), add = TRUE)
+  ## FIXME: OpenMP reporting is not quite right
+  ## n = 1L rather than NULL so we report "setting" rather than "current"
+  ## autopar switches from NULL to FALSE (default)  
+  parallel <- object$modelInfo$parallel
+  if (!all(sapply(parallel, is.null))) {
+      n_orig <- openmp()
+      do.call(openmp, parallel)
+      on.exit(do.call(openmp, n_orig), add = TRUE)
   }
 
-  if (openmp_debug()) cat("TMB threads currently set to ", openmp(NULL), "\n")
-  return_eta <- type %in% c("zlink", "link")
+
+  if (openmp_debug()) {
+      cat("TMB threads currently set to ", openmp(NULL), "\n")
+  }
+  return_par <- if (type %in% c("zlink", "link")) "eta_predict" else if (type=="latent") "b" else "mu_predict"
+
   if (!se.fit) {
     rr <- newObj$report(lp)
-    pred <- if (return_eta) rr$eta_predict else rr$mu_predict
+    pred <- rr[[return_par]]
   } else {
     H <- with(object,optimHess(oldPar,obj$fn,obj$gr))
     ## FIXME: Eventually add 'getReportCovariance=FALSE' to this sdreport
@@ -451,15 +499,19 @@ predict.glmmTMB <- function(object,
     if (cov.fit) {
         sdr <- sdreport(newObj,oldPar,hessian.fixed=H,getReportCovariance=TRUE)
         covfit <- sdr$cov
-    } else sdr <- sdreport(newObj,oldPar,hessian.fixed=H,getReportCovariance=FALSE)
+    } else     sdr <- sdreport(newObj,oldPar,hessian.fixed=H,getReportCovariance=FALSE,bias.correct=do.bias.correct,bias.correct.control=bias.correct.control)
+
     sdrsum <- summary(sdr, "report") ## TMB:::summary.sdreport(sdr, "report")
-    w <- if (return_eta) "eta_predict" else "mu_predict"
-    ## multiple rows with identical names; naive indexing
-    ## e.g. sdrsum["mu_predict", ...] returns only the first instance
-    w <- which(rownames(sdrsum)==w)
-    pred <- sdrsum[w,"Estimate"]
-    se <- sdrsum[w,"Std. Error"]
+    ## split summary matrix by parameter name
+    sdrsplit <- split.data.frame(sdrsum, rownames(sdrsum))
+    pred <- sdrsplit[[return_par]][,"Estimate"]
+    se <- sdrsplit[[return_par]][,"Std. Error"]
+    w <- which(rownames(sdrsum) == return_par)
     if (cov.fit) covfit <- covfit[w, w]
+    
+    if (do.bias.correct) {
+        return (sdrsum[w,])
+    }
   }
   if (do.napred) {
       pred <- napredict(na.act,pred)
