@@ -6,7 +6,7 @@ stopifnot(require("testthat"),
 drop_version <- function(obj) {
     obj$modelInfo$packageVersion <- NULL
     obj$modelInfo$family$initialize <- NULL  ## updated initialization expressions
-    obj
+    obj$modelInfo$parallel <- NULL   ## parallel component changed from int to list
 }
 
 expect_equal_nover <- function(x,y,...) {
@@ -155,9 +155,9 @@ test_that("Update Binomial", {
 })
 
 test_that("internal structures", {
-  ## RE terms only in cond and zi model, not disp: GH #79
+  ## RE terms in cond, zi, and disp model
   expect_equal(names(fm0$modelInfo$reTrms),
-               c("cond","zi"))
+               c("cond","zi", "disp"))
 })
 
 test_that("close to lme4 results", {
@@ -379,3 +379,26 @@ test_that("handle failure in numDeriv::jacobian",
                 family = truncated_nbinom1))
    expect_is(m1, "glmmTMB")
 })
+
+test_that("don't mess up internal formula ordering", {
+    dd <- expand.grid(type = factor(c("R","S")),
+                      strainN = factor(1:3),
+                      dose = c("N","C"),
+                      rep = factor(1:5),
+                      aid = factor(1:5),
+                      rep = 1:10)
+    set.seed(101)
+    dd$l3 <- rnorm(nrow(dd))
+    form <- l3~(type/strainN)*dose
+    m1 <- glmmTMB(form, data = dd)
+    expect_identical(colnames(model.matrix(m1)),
+                     colnames(model.matrix(form, data = dd)))
+})
+
+test_that("subset argument", {
+    fms <- update(fm1, subset = Days > 0)
+    expect_false(isTRUE(all.equal(fixef(fm1), fixef(fms))))
+    expect_equal(fixef(fms)$cond,
+                 c(`(Intercept)` = 248.63573143019875,
+                   Days = 10.904528975303338))
+})
