@@ -417,7 +417,7 @@ test_that("tweedie", {
     expect_equal(fixef(twm)$cond, fixef(twm2)$cond, tolerance = 1e-1)
     expect_equal(sigma(twm), sigma(twm2), tolerance = 1e-1)
     expect_equal(ranef(twm),
-                 structure(list(cond = list(), zi = list()), class = "ranef.glmmTMB"))
+                 structure(list(cond = list(), zi = list(), disp = list()), class = "ranef.glmmTMB"))
 })
 
 test_that("gaussian_sqrt", {
@@ -498,3 +498,85 @@ test_that("t-distributed response", {
                  tolerance = 1e-6)
 })
 
+test_that("nbinom12 family", {
+    set.seed(101)
+    n <- 10000
+    x <- rnorm(n)
+    mu <- exp(2 + 1*x)
+    vv <- mu*(1+2+mu/0.5)
+    k <- mu/(vv/mu - 1)
+    dd <- data.frame(x, y = rnbinom(n, mu = mu, size = k))
+    m1 <- glmmTMB(y ~ x, family = nbinom12, data = dd)
+    ## basic test
+    ## should have phi = 2, k = 0.5
+    ## log(phi) ~ 0.7, log(psi) ~ -0.7
+    expect_equal(    m1$obj$env$last.par.best,
+                 c(beta = 1.98948426828242, beta = 1.00635151325394,
+                   betadisp = 0.68344614610532, psi = -0.686823594633112),
+                 tolerance = 1e-6)
+    expect_equal(sigma(m1), 1.980692, tolerance = 1e-6)
+})
+
+test_that("skewnormal family", {
+    dd <- data.frame(dummy = rep(1, 500))
+    dd$y <- simulate_new(~1,
+                            newdata = dd,
+                            newparams = list(beta = -1,
+                                             betadisp = 3,
+                                             psi = -5),
+                            seed = 101,
+                            family = "skewnormal")[[1]]
+    expect_equal(range(dd$y), c(-64.8363758099827, 32.87734399648))
+    expect_equal(length(unique(dd$y)), 500L)
+    fit <- glmmTMB(y ~ 1,
+                   data = dd,
+                   family = "skewnormal",
+                   start = list(betadisp = log(sd(dd$y)),
+                                psi = -5))
+    expect_equal(fit$obj$env$last.par.best,
+                 c(beta = 0.0765490512716489,
+                   betadisp = 2.94927708520387,
+                   psi = -6.12362878509844),
+                 tolerance = 1e-6)
+    expect_equal(family_params(fit),
+                 c(`Skewnormal shape` = -6.12362878509844),
+                 tolerance = 1e-6)
+})
+
+test_that("make_family initialize works", {
+    ## GH #1133
+    if (require(effects)) {
+        data("sleepstudy", package = "lme4")
+        m2 <- glmmTMB(round(Reaction) ~ Days + (1 | Subject), sleepstudy,
+                      family = truncated_nbinom2)
+        ee <- suppressWarnings(effects::Effect("Days", m2))
+        expect_equal(c(ee$fit),
+                     c(5.5317435373068, 5.6003872162399,
+                       5.669030895173, 5.77199641357265, 
+                       5.84064009250575))
+    }
+})
+
+test_that("testing family specification", {
+    ## test S4 (motivating example was VGAM::zipoisson,
+    ##  but we want an S4 object that comes from packages we already
+    ## depend on
+
+    ## family() returns S4 object
+    expect_error(glmmTMB(hp ~ mpg, data = mtcars,
+                         family = Matrix::Matrix),
+                 "must be a list")
+    ## family() returns list without $family element
+    expect_error(glmmTMB(hp ~ mpg, data = mtcars,
+                         family = function() list()),
+                 "must be a list")
+    ## family is a non-existent object
+    expect_error(glmmTMB(hp ~ mpg, data = mtcars,
+                         family = zipoisson),
+                 "not found")
+    ## get(family) doesn't find a function
+    expect_error(glmmTMB(hp ~ mpg, data = mtcars,
+                         family = "zipoisson"),
+                 "of mode.*not found")
+
+})
